"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const inversify_1 = require("inversify");
const lib_reflect_1 = require("lib-reflect");
const lodash_1 = require("lodash");
const activation_execution_1 = require("./activation.execution");
const execute_1 = require("./activations/execute");
const execute_static_1 = require("./activations/execute.static");
const type_param_inject_1 = require("./activations/type.param.inject");
const singleton_1 = require("./attributes/singleton");
const interceptor_data_1 = require("./interceptor.data");
class ActivationsGenerator {
    constructor() {
        this._classes = [];
        this._allActivations = [];
    }
    register(target) {
        const registeredClass = this._classes.find((classData) => classData.target === target);
        if (!lodash_1.isNil(registeredClass)) {
            throw new Error(`Cannot register class '${target.name}' for activation, already registered.`);
        }
        const reflectData = lib_reflect_1.ReflectHelper.getClass(target);
        if (lodash_1.isNil(reflectData)) {
            throw new Error(`Cannot register class '${target.name}' for activation, class is not found in reflection system.`);
        }
        reflectData.build();
        this._classes.push(reflectData);
        return this;
    }
    addActivations(interceptor, type) {
        const interceptorData = new interceptor_data_1.InterceptorData();
        interceptorData.target = interceptor;
        if (!lodash_1.isNil(type)) {
            interceptorData.type = type;
        }
        interceptorData.createInstance();
        if (!this._isBeforeActivation(interceptorData.instance) && !this._isAfterActivation(interceptorData.instance)) {
            throw new Error(`Object is not an interceptor '${interceptorData.instance.constructor.name}'.`);
        }
        this._allActivations.push(interceptorData);
        return this;
    }
    generateActivations(container) {
        // For each class, for each method, generate the activations
        const activations = [];
        lodash_1.forEach(this._classes, (classData) => {
            // Check if the class is decorated with inject
            if (!Reflect.hasOwnMetadata(inversify_1.METADATA_KEY.PARAM_TYPES, classData.target)) {
                inversify_1.injectable()(classData.target);
            }
            if (!container.isBound(classData.target)) {
                const boundInstance = container.bind(classData.target).toSelf();
                // Singleton?
                if (classData.tags[singleton_1.kInterceptorSingletonKey]) {
                    boundInstance.inSingletonScope();
                }
            }
            // For each method on those classes
            lodash_1.forEach(classData.methods, (methodData) => {
                const activation = new activation_execution_1.Activation();
                activation.class = classData;
                activation.method = methodData;
                activation.beforeActivation = [];
                activation.afterActivation = [];
                activations.push(activation);
            });
        });
        // The final interceptor (the one that actually executes the function call)
        const executeInterceptor = new interceptor_data_1.InterceptorData();
        executeInterceptor.target = execute_1.ExecuteActivation;
        executeInterceptor.createInstance();
        const executeInterceptorStatic = new interceptor_data_1.InterceptorData();
        executeInterceptorStatic.target = execute_static_1.ExecuteActivationStatic;
        executeInterceptorStatic.createInstance();
        // Injector for Type parameters registered in the container
        const typeInterceptor = new interceptor_data_1.InterceptorData();
        typeInterceptor.target = type_param_inject_1.TypeParamInterceptor;
        typeInterceptor.createInstance();
        // Create the activations list
        lodash_1.forEach(activations, (activation) => {
            // Go through all activations and check if we add them in the before / after list
            this._addActivations(activation, this._allActivations);
            // Now add the class activations if any
            const classActivations = activation.class.getAttributesOfType(interceptor_data_1.InterceptorData);
            if (!lodash_1.isEmpty(classActivations)) {
                this._addActivations(activation, classActivations.reverse());
            }
            // Now add the method activations if any
            const methodActivations = activation.method.getAttributesOfType(interceptor_data_1.InterceptorData);
            if (!lodash_1.isEmpty(methodActivations)) {
                this._addActivations(activation, methodActivations.reverse());
            }
            activation.beforeActivation.push(typeInterceptor.instance);
            if (activation.method.flags === lib_reflect_1.MethodFlags.STATIC) {
                // Finally add the execution interceptor
                activation.beforeActivation.push(executeInterceptorStatic.instance);
            }
            else {
                // Finally add the execution interceptor
                activation.beforeActivation.push(executeInterceptor.instance);
            }
        });
        return activations;
    }
    _isBeforeActivation(instance) {
        return lodash_1.isFunction(instance.before);
    }
    _isAfterActivation(instance) {
        return lodash_1.isFunction(instance.after);
    }
    _addActivations(activation, activations) {
        lodash_1.forEach(activations, (interceptor) => {
            interceptor.createInstance();
            const interceptorInstance = interceptor.instance;
            if ((interceptor.type & interceptor_data_1.InterceptorType.Before) === interceptor_data_1.InterceptorType.Before &&
                this._isBeforeActivation(interceptorInstance)) {
                activation.beforeActivation.push(interceptorInstance);
            }
            if ((interceptor.type & interceptor_data_1.InterceptorType.After) === interceptor_data_1.InterceptorType.After &&
                this._isAfterActivation(interceptorInstance))
                // It must be an After processor
                activation.afterActivation.push(interceptorInstance);
        });
    }
}
exports.ActivationsGenerator = ActivationsGenerator;
//# sourceMappingURL=activation.generator.js.map